import typing
from pathlib import Path
from typing import Optional

from sema4ai.actions import Row

if typing.TYPE_CHECKING:
    import urllib3

    from sema4ai.data._result_set import ResultSet

from robocorp import log


class _HttpConnectionHelper:
    """
    Helper class to manage connections to the http server.
    """

    def __init__(self, http_url: str, http_user: str, http_password: str) -> None:
        self._http_url = http_url
        self._http_user = http_user
        self._http_password = http_password
        self._session_headers: dict[str, str] = {}

    def login(self) -> None:
        import time
        from http.cookies import SimpleCookie

        import sema4ai_http

        login_url = f"{self._http_url}/api/login"

        login_response = sema4ai_http.post(
            login_url,
            json={"password": self._http_password, "username": self._http_user},
        )

        if login_response.status != 200:
            # 401 is Unauthorized (bad or expired credentials, no need to retry)
            if login_response.status != 401:
                # Retry once in case of another unexpected error
                time.sleep(0.2)
                login_response = sema4ai_http.post(
                    login_url,
                    json={"password": self._http_password, "username": self._http_user},
                )

            if login_response.status != 200:
                raise Exception(
                    f"Failed to login. Status: {login_response.status}. Data: {login_response.data.decode('utf-8', errors='backslashreplace')}"
                )

        cookies = SimpleCookie()
        session_cookies = {}
        if "set-cookie" in login_response.headers:
            cookies.load(login_response.headers["set-cookie"])
            session_cookies = {key: morsel.value for key, morsel in cookies.items()}
        cookie_header = "; ".join([f"{k}={v}" for k, v in session_cookies.items()])
        self._session_headers = {"Cookie": cookie_header}

    def _is_login_info_available(self) -> bool:
        return bool(self._http_user and self._http_password)

    def _login_if_needed(self) -> None:
        if not self._session_headers and self._is_login_info_available():
            self.login()

    def upload_file(self, file_path: Path, table_name: str) -> None:
        self._login_if_needed()
        try:
            self._upload_file(file_path, table_name)
        except Exception:
            if not self._is_login_info_available():
                raise

            # Retry once with a new session
            self._session_headers = {}
            self._login_if_needed()
            self._upload_file(file_path, table_name)

    def _run_sql_with_retry(
        self, sql: str, database: str | None = None
    ) -> "urllib3.BaseHTTPResponse":
        import time

        import sema4ai_http

        url = self._http_url + "/api/sql/query"

        result = sema4ai_http.post(
            url,
            json={"query": sql, "context": {"db": database or ""}},
            headers=self._session_headers,
        )

        if result.status != 200:
            if result.status == 401:  # Unauthorized (bad or expired credentials)
                # Try to login again
                self._session_headers = {}
                self._login_if_needed()
            else:
                # Another unexpected error (sleep a bit and retry)
                time.sleep(0.2)

            # Retry the request (once)
            result = sema4ai_http.post(
                url,
                json={"query": sql, "context": {"db": database or ""}},
                headers=self._session_headers,
            )

        return result

    def run_sql(
        self, sql: str, database: str | None = None
    ) -> tuple[list[str], list[Row]] | None:
        """
        Args:
            sql: The SQL query to execute.
            database: The database to use.

        Returns:
            A tuple of columns and rows or None if a table was not returned.
        """
        from sema4ai.actions import ActionError

        self._login_if_needed()

        result = self._run_sql_with_retry(sql, database)

        if result.status != 200:
            raise Exception(
                f"Failed to run sql. Status: {result.status}. Data: {result.data.decode('utf-8', errors='backslashreplace')}"
            )

        data = result.json()
        data_type = data["type"]
        if data_type == "table":
            columns: list[str] = [x for x in data["column_names"]]
            rows: list[Row] = data["data"]
            return columns, rows
        if data_type == "ok":
            return None
        if data_type == "error":
            log.critical(
                f"There was an error running the SQL query: {sql}.\n"
                "Please verify that the tables being queried are prefixed by the Data Source name and that the Data Source name has the appropriate ` escaping.\n"
                "Example: `SELECT * FROM `my-datasource-name`.table_name`.\n"
                f"Original error: {data['error_message']}"
            )
            raise ActionError("There was an error running the query SQL.")

        log.critical(f"Unexpected sql result type: {data_type}")
        return None

    def _upload_file(self, file_path: Path, table_name: str) -> None:
        import json

        import sema4ai_http

        file_name = file_path.name
        data = file_path.read_bytes()
        result = sema4ai_http.put(
            f"{self._http_url}/api/files/{table_name}",
            fields={
                "file": (file_name, data),
                "data": json.dumps(
                    {
                        "original_file_name": file_name,
                        "name": table_name,
                        "source_type": "file",
                    }
                ).encode("utf-8"),
            },
            headers=self._session_headers,
        )
        if result.status != 200:
            raise Exception(
                f"Failed to upload file. Status: {result.status}. Data: {result.data.decode('utf-8', errors='backslashreplace')}"
            )


class DataServerConnection:
    def __init__(
        self,
        http_url: str,
        http_user: Optional[str],
        http_password: Optional[str],
        mysql_host: str,
        mysql_port: int,
        mysql_user: Optional[str],
        mysql_password: Optional[str],
    ):
        """
        Creates a connection to the data server.
        """
        # Not using mysql connection for now (had issues with pymysql not giving
        # errors when the connection was closed by the server).
        self._http_connection = _HttpConnectionHelper(
            http_url, http_user or "", http_password or ""
        )

    def query(
        self,
        query: str,
        params: Optional[dict[str, str | int | float] | list[str | int | float]] = None,
    ) -> "ResultSet":
        """
        Simple API to query a database in MindsDB with parameters. Always loads
        all the results into memory.

        Args:
            query: The SQL query to execute.
            params: A list of parameters to inject into the query.

        Returns:
            ResultSet: The query result as a ResultSet.
        """
        from sema4ai.data._result_set import ResultSet
        from sema4ai.data._sql_handling import (
            build_query_from_dict_params,
            build_query_from_list_params,
        )

        if isinstance(params, list):
            query = build_query_from_list_params(query, params)
        else:
            query = build_query_from_dict_params(query, params)

        # For now, we don't support database names (scope is always global and
        # users need to specify the full data source name + table name in the
        # SQL being executed).
        database_name = ""
        result = self._http_connection.run_sql(query, database_name)
        if result is None:
            raise RuntimeError(
                "Unexpected result from the data server (expected table but received just 'ok')"
            )

        columns, rows = result
        return ResultSet(columns, rows)

    # It's actually the same thing internally, so we can just alias it.
    predict = query



from pathlib import Path
from typing import Any, Callable, Optional, overload

from ._data_server_connection import DataServerConnection
from ._data_source import DataSource
from ._data_source_spec import DataSourceSpec
from ._result_set import ResultSet

__version__ = "0.0.4"
version_info = [int(x) for x in __version__.split(".")]


@overload
def query(func: Callable) -> Callable:
    ...


@overload
def query(
    *, is_consequential: Optional[bool] = None, display_name: Optional[str] = None
) -> Callable:
    ...


def query(*args, **kwargs):
    """
    Decorator for queries which can be executed by `sema4ai.actions`.

    i.e.:

    If a file such as actions.py has the contents below:

    ```python
    from sema4ai.data import query

    @query
    def query_invoices(date: str, datasource: DataSource[Literal["invoices-datasource"]]) -> list[str]:
        ...
    ```

    Note that a query needs sema4ai actions to be executed.
    The command line to execute it is:

    python -m sema4ai.actions run actions.py -a query_invoices

    Args:
        func: A function which is a query to `sema4ai.data`.
        is_consequential: Whether the action is consequential or not.
            This will add `x-openai-isConsequential: true` to the action
            metadata and shown in OpenApi spec.
        display_name: A name to be displayed for this action.
            If given will be used as the openapi.json summary for this action.
    """
    from sema4ai.actions import action

    kwargs["kind"] = "query"
    return action(*args, **kwargs)


@overload
def predict(
    *, is_consequential: Optional[bool] = None, display_name: Optional[str] = None
) -> Callable:
    ...


@overload
def predict(func: Callable) -> Callable:
    ...


def predict(*args, **kwargs):
    """
    Decorator for predictions based on data models which can be executed by `sema4ai.actions`.

    i.e.:

    If a file such as actions.py has the contents below:

    ```python
    from sema4ai.data import predict

    @predict
    def cashflow_prediction(scenario: str, datasource: DataSource[Literal["reports-datasource"]]) -> list[str]:
        ...
    ```

    Note that a prediction needs sema4ai actions to be executed.
    The command line to execute it is:

    python -m sema4ai.actions run actions.py -a cashflow_prediction

    Args:
        func: A function which is a prediction to `sema4ai.data`.
        is_consequential: Whether the action is consequential or not.
            This will add `x-openai-isConsequential: true` to the action
            metadata and shown in OpenApi spec.
        display_name: A name to be displayed for this action.
            If given will be used as the openapi.json summary for this action.
    """
    from sema4ai.actions import action

    kwargs["kind"] = "predict"
    return action(*args, **kwargs)


def metadata(package_root: Path) -> dict:
    """
    Returns something like this:
    {
        // This is added by sema4ai-data
        'data_spec_version': 'v2',
        'data': {
            'datasources': [
                {
                    'name': 'datasource-name',
                    'engine': 'sqlite',
                    'description': 'description of the datasource',
                    'setup_sql': ['sql-to-execute'],
                    'defined_at': {
                        'file': 'relative/path/to/definition_file.py',
                        'line': 123,
                    },

                    # Only for files engine
                    'file': 'path/to/definition_file.py',  # Relative to package root.
                    'created_table': 'name-of-created-table',

                    # Only if a model is created (prediction engine or custom engine)
                    'model_name': 'name-of-model',

                },
                ...
            ]
        },
    }
    """
    from sema4ai.data._data_source_spec import (
        _global_internal_data_server_datasources_spec_instances,
    )

    ret: dict[str, Any] = {}
    datasources: list[dict[str, Any]] = []
    for (
        dataserver_datasource_spec
    ) in _global_internal_data_server_datasources_spec_instances:
        datasources.append(dataserver_datasource_spec.get_metadata(package_root))

    ret["data"] = {"datasources": datasources}
    ret["data_spec_version"] = "v2"
    return ret


def get_connection() -> "DataServerConnection":
    """
    Returns a connection to the data server.

    Usually it's recommended to use the `DataSource` as a parameter in the @query/@predict,
    but if needed, this method can also be used to get the configured connection.
    """
    from sema4ai.data._data_source import ConnectionNotSetupError

    from ._data_source import _ConnectionHolder

    try:
        return _ConnectionHolder.connection()
    except ConnectionNotSetupError as e:
        # Let's see if we can auto-setup the connection now.

        # This is not public for external use, but from internal use it's ok for now.
        from sema4ai.actions._action import get_current_requests_contexts

        ctx = get_current_requests_contexts()
        if not ctx:
            # We can't auto-setup the connection if there's no context.
            raise e

        data_context = ctx.data_context
        if data_context:
            DataSource.setup_connection_from_data_context(data_context)
            try:
                return _ConnectionHolder.connection()
            except ConnectionNotSetupError:
                pass

        # If it's still not setup, try to setup from env vars.
        DataSource.setup_connection_from_env_vars()
        try:
            return _ConnectionHolder.connection()
        except ConnectionNotSetupError:
            pass

        # Unable to auto-setup the connection, just raise the original error.
        raise e


__all__ = [
    "query",
    "predict",
    "metadata",
    "DataSource",
    "ResultSet",
    "DataSourceSpec",
    "get_connection",
]



import typing
from pathlib import Path
from typing import Any, Literal, Optional

_global_internal_data_server_datasources_spec_instances: list["DataSourceSpec"] = []


JSONValue = typing.Union[
    dict[str, "JSONValue"], list["JSONValue"], list[str], str, int, float, bool, None
]


def _get_datasource_metadata_from_annotation_args(tp) -> dict[str, Any]:
    from sema4ai.actions._exceptions import ActionsCollectError

    error_message = """
Invalid DataSource annotation found.

The DataSource must always be annotated with a DataSourceSpec:
(i.e.: `Annotated[DataSource, DataSourceSpec(name="datasource_name", engine="postgres")]`) 
"""
    annotation_args = typing.get_args(tp)
    if not annotation_args or len(annotation_args) != 2:
        raise ActionsCollectError(error_message)

    datasource_spec = annotation_args[1]
    if datasource_spec.__class__.__name__ != "DataSourceSpec":
        raise ActionsCollectError(
            f"Second parameter in Annotation of DataSource is not a DataSourceSpec.\n{error_message}"
        )

    return datasource_spec.get_metadata()


_CUSTOM_ENGINE_NAMES = (
    "prediction:lightwood",  # Special use case: we know this maps to a lightwood prediction model
    "custom",  # Anything we don't really expect (that can be configured with custom SQL) should be here
    "mindsdb",  # Special use case: we know this maps to the `mindsdb` pre-configured project.
    # All the other engines are supported by the data server and are queried with:
    # SHOW HANDLERS WHERE type = 'data'
)

_DATASOURCES_ENGINE_NAMES = (
    "access",
    "aerospike",
    "airtable",
    "altibase",
    "apache_doris",
    "aqicn",
    "athena",
    "aurora",
    "bigquery",
    "binance",
    "cassandra",
    "chromadb",
    "ckan",
    "clickhouse",
    "cloud_spanner",
    "cloud_sql",
    "cockroachdb",
    "coinbase",
    "confluence",
    "couchbasevector",
    "couchbase",
    "crate",
    "d0lt",
    "databend",
    "databricks",
    "astra",
    "db2",
    "derby",
    "discord",
    "dockerhub",
    "documentdb",
    "dremio",
    "druid",
    "duckdb",
    "dummy_data",
    "dynamodb",
    "edgelessdb",
    "elasticsearch",
    "email",
    "empress",
    "eventbrite",
    "eventstoredb",
    "faunadb",
    "files",
    "financial_modeling_prep",
    "firebird",
    "frappe",
    "gcs",
    "github",
    "gitlab",
    "gmail",
    "google_analytics",
    "google_books",
    "google_calendar",
    "google_content_shopping",
    "google_fit",
    "google_search",
    "greptimedb",
    "hackernews",
    "hana",
    "hive",
    "hsqldb",
    "hubspot",
    "ignite",
    "impala",
    "influxdb",
    "informix",
    "ingres",
    "instatus",
    "intercom",
    "jira",
    "kinetica",
    "lancedb",
    "libsql",
    "lightdash",
    "lindorm",
    "luma",
    "mariadb",
    "materialize",
    "matrixone",
    "maxdb",
    "mediawiki",
    "mendeley",
    "milvus",
    "monetdb",
    "mongodb",
    "mssql",
    "teams",
    "mysql",
    "newsapi",
    "notion",
    "npm",
    "nuo_jdbc",
    "oceanbase",
    "oilpriceapi",
    "openbb",
    "opengauss",
    "openstreetmap",
    "oracle",
    "orioledb",
    "paypal",
    "pgvector",
    "phoenix",
    "pinecone",
    "pinot",
    "pirateweather",
    "plaid",
    "planet_scale",
    "postgres",
    "pypi",
    "qdrant",
    "questdb",
    "quickbooks",
    "reddit",
    "redshift",
    "rocket_chat",
    "rockset",
    "s3",
    "salesforce",
    "sap_erp",
    "scylladb",
    "sendinblue",
    "serpstack",
    "sharepoint",
    "sheets",
    "shopify",
    "singlestore",
    "slack",
    "snowflake",
    "solace",
    "solr",
    "sqlany",
    "sqlite",
    "sqreamdb",
    "starrocks",
    "strapi",
    "strava",
    "stripe",
    "supabase",
    "surrealdb",
    "symbl",
    "tdengine",
    "teradata",
    "tidb",
    "timescaledb",
    "trino",
    "tripadvisor",
    "twilio",
    "twitter",
    "vertica",
    "vitess",
    "weaviate",
    "webz",
    "web",
    "whatsapp",
    "xata",
    "youtube",
    "yugabyte",
    "zendesk",
    "zipcodebase",
    "zotero",
)

_VALID_ENGINE_NAMES = _CUSTOM_ENGINE_NAMES + _DATASOURCES_ENGINE_NAMES


class DataSourceSpec:
    def __init__(
        self,
        *,
        engine: str,
        name: Optional[
            Literal[
                "prediction:lightwood",  # Special use case: we know this maps to a lightwood prediction model
                "custom",  # Anything we don't really expect (that can be configured with custom SQL) should be here
                "models",  # Special use case: we know this maps to the `models` pre-configured data source.
                "files",  # Special use case: we know this maps to the `files` pre-configured data source.
                # All the other engines are supported by the data server and are queried with:
                # SHOW HANDLERS WHERE type = 'data'
                "access",
                "aerospike",
                "airtable",
                "altibase",
                "apache_doris",
                "aqicn",
                "athena",
                "aurora",
                "bigquery",
                "binance",
                "cassandra",
                "chromadb",
                "ckan",
                "clickhouse",
                "cloud_spanner",
                "cloud_sql",
                "cockroachdb",
                "coinbase",
                "confluence",
                "couchbasevector",
                "couchbase",
                "crate",
                "d0lt",
                "databend",
                "databricks",
                "astra",
                "db2",
                "derby",
                "discord",
                "dockerhub",
                "documentdb",
                "dremio",
                "druid",
                "duckdb",
                "dummy_data",
                "dynamodb",
                "edgelessdb",
                "elasticsearch",
                "email",
                "empress",
                "eventbrite",
                "eventstoredb",
                "faunadb",
                "financial_modeling_prep",
                "firebird",
                "frappe",
                "gcs",
                "github",
                "gitlab",
                "gmail",
                "google_analytics",
                "google_books",
                "google_calendar",
                "google_content_shopping",
                "google_fit",
                "google_search",
                "greptimedb",
                "hackernews",
                "hana",
                "hive",
                "hsqldb",
                "hubspot",
                "ignite",
                "impala",
                "influxdb",
                "informix",
                "ingres",
                "instatus",
                "intercom",
                "jira",
                "kinetica",
                "lancedb",
                "libsql",
                "lightdash",
                "lindorm",
                "luma",
                "mariadb",
                "materialize",
                "matrixone",
                "maxdb",
                "mediawiki",
                "mendeley",
                "milvus",
                "monetdb",
                "mongodb",
                "mssql",
                "teams",
                "mysql",
                "newsapi",
                "notion",
                "npm",
                "nuo_jdbc",
                "oceanbase",
                "oilpriceapi",
                "openbb",
                "opengauss",
                "openstreetmap",
                "oracle",
                "orioledb",
                "paypal",
                "pgvector",
                "phoenix",
                "pinecone",
                "pinot",
                "pirateweather",
                "plaid",
                "planet_scale",
                "postgres",
                "pypi",
                "qdrant",
                "questdb",
                "quickbooks",
                "reddit",
                "redshift",
                "rocket_chat",
                "rockset",
                "s3",
                "salesforce",
                "sap_erp",
                "scylladb",
                "sendinblue",
                "serpstack",
                "sharepoint",
                "sheets",
                "shopify",
                "singlestore",
                "slack",
                "snowflake",
                "solace",
                "solr",
                "sqlany",
                "sqlite",
                "sqreamdb",
                "starrocks",
                "strapi",
                "strava",
                "stripe",
                "supabase",
                "surrealdb",
                "symbl",
                "tdengine",
                "teradata",
                "tidb",
                "timescaledb",
                "trino",
                "tripadvisor",
                "twilio",
                "twitter",
                "vertica",
                "vitess",
                "weaviate",
                "webz",
                "web",
                "whatsapp",
                "xata",
                "youtube",
                "yugabyte",
                "zendesk",
                "zipcodebase",
                "zotero",
            ]
        ] = None,
        created_table: Optional[str] = None,
        model_name: Optional[str] = None,
        file: Optional[str] = None,
        description: str = "",
        setup_sql: Optional[str | list[str]] = None,
        setup_sql_files: Optional[str | list[str]] = None,
    ):
        import sys

        if not engine:
            raise ValueError(
                "'engine' of the datasource must be set (some common values: 'postgres', 'files', 'prediction:lightwood', 'redshift')"
            )

        if engine == "files":
            if not name:
                name = "files"

        if engine.startswith("prediction:") or name == "custom":
            if not name:
                name = "models"

        self.name = name
        self.engine = engine
        self.created_table = created_table
        self.model_name = model_name
        self.file = file
        self.description = description
        self.setup_sql = setup_sql
        frame = sys._getframe(1)
        self.__defined_at_file = frame.f_code.co_filename
        self.__defined_at_line = frame.f_lineno
        self.setup_sql_files = setup_sql_files

        self._validate()

        _global_internal_data_server_datasources_spec_instances.append(self)

    def _validate(self):
        import os.path

        if not self.name:
            raise ValueError("'name' of the datasource must be set")

        if not self.engine:
            raise ValueError(
                "'engine' of the datasource must be set (some common values: 'postgres', 'files', 'prediction:lightwood', 'redshift')"
            )

        if not self.description:
            raise ValueError("'description' of the datasource must be set")

        if self.engine == "files":
            if self.name != "files":
                raise ValueError("'name' must be 'files' for the 'files' engine")

            if not self.created_table:
                raise ValueError("'created_table' must be set for the 'files' engine")

            if not self.file:
                raise ValueError("'file' must be set for the 'files' engine")

            if os.path.isabs(self.file):
                raise ValueError("'file' must be a relative path")

        if self.engine not in _VALID_ENGINE_NAMES:
            raise ValueError(
                f"Invalid engine: {self.engine} (note: consider using 'custom' if it's not a pre-defined supported engine and it's configured with SQL). Valid engines are: {_VALID_ENGINE_NAMES}"
            )

        if self.engine in _CUSTOM_ENGINE_NAMES:
            if not self.setup_sql and not self.setup_sql_files:
                raise ValueError(
                    f"The engine: {self.engine} requires a 'sql' to be set (which will be executed to create the datasource/table)"
                )
            if self.setup_sql and self.setup_sql_files:
                raise ValueError(
                    "It's not possible to use both 'setup_sql' and 'setup_sql_files' parameters together"
                )

            if self.setup_sql_files:
                if not isinstance(self.setup_sql_files, list):
                    raise ValueError(
                        f"'setup_sql_files' must be a list of file paths (relative to the package root). Found: {type(self.setup_sql_files)}"
                    )

                for setup_sql_file in self.setup_sql_files:
                    if os.path.isabs(setup_sql_file):
                        raise ValueError(
                            f"'setup_sql_files' must contain relative paths (relative to the package root). Found absolute path: {setup_sql_file}"
                        )
        else:
            if self.setup_sql or self.setup_sql_files:
                raise ValueError(
                    f"The engine: {self.engine} does not support the 'sql' parameter (only {_CUSTOM_ENGINE_NAMES} engines support it)"
                )

        if self.engine.startswith("prediction:"):
            if not self.model_name:
                raise ValueError(
                    f"The engine: {self.engine} requires a 'model_name' to be set"
                )

        if self.created_table:
            for char in ["'", '"', "`", ";"]:
                if char in self.created_table:
                    raise ValueError(
                        f"created_table name must not contain character: {char}"
                    )
        if self.model_name:
            for char in ["'", '"', "`", ";"]:
                if char in self.model_name:
                    raise ValueError(f"model_name must not contain character: {char}")

    def get_metadata(self, package_root: Path) -> dict[str, JSONValue]:
        try:
            defined_at_file = (
                Path(self.__defined_at_file).relative_to(package_root).as_posix()
            )
        except Exception:
            defined_at_file = Path(self.__defined_at_file).as_posix()

        ret: dict[str, JSONValue] = {
            "name": self.name,
            "engine": self.engine,
            "description": self.description,
            "defined_at": {
                "file": defined_at_file,
                "line": self.__defined_at_line,
            },
        }
        if self.created_table:
            ret["created_table"] = self.created_table
        if self.model_name:
            ret["model_name"] = self.model_name
        if self.file:
            ret["file"] = self.file
        if self.setup_sql:
            if isinstance(self.setup_sql, str):
                ret["setup_sql"] = [self.setup_sql]
            else:
                assert isinstance(self.setup_sql, list)
                ret["setup_sql"] = self.setup_sql
        if self.setup_sql_files:
            # If we have a setup_sql_files, we'll read those files and use the content as setup_sql
            sqls: list[str] = []
            for setup_sql_file in self.setup_sql_files:
                setup_sql_files_path = package_root / setup_sql_file
                try:
                    sqls.append(Path(setup_sql_files_path).read_text(encoding="utf-8"))
                except Exception:
                    raise ValueError(
                        f"Failed to read setup_sql_files content from: {setup_sql_file} -- resolved path: {setup_sql_files_path}"
                    )
            ret["setup_sql"] = sqls

        return ret



from typing import Optional


def build_query_from_dict_params(
    query: str, params: Optional[dict[str, str | int | float]]
) -> str:
    """
    Replace all the $name in the query with the actual values from params.

    Args:
        query: The SQL query with placeholders.
        params: A dictionary of parameters to inject into the query.

    Returns:
        str: The query with parameters injected.
    """
    import math
    import re

    if not params and "$" not in query:
        return query

    new_params: dict[str, str] = {}
    if params:
        # Escape single quotes in parameters to prevent SQL injection
        C1 = "'"
        C2 = "''"
        for key, value in params.items():
            if isinstance(value, str):
                value = value.replace(C1, C2)
                new_params[key] = f"'{value}'"
            elif isinstance(value, (int, float)):
                if not math.isfinite(value):
                    raise ValueError(
                        f"Infinity or NaN is not supported as a query parameter. Parameters: {params}"
                    )
                new_params[key] = str(value)
            else:
                raise ValueError(f"Unsupported parameter: {value} type: {type(value)}")

    # Replace placeholders with actual values
    def replace_placeholder(match):
        key = match.group(1)
        if key in new_params:
            return new_params[key]
        raise ValueError(f"Missing parameter for placeholder: {key}")

    # Use regex to find all placeholders and replace them
    query = re.sub(r"\$(\w+)", replace_placeholder, query)

    return query


def build_query_from_list_params(
    query: str, params: Optional[list[str | int | float]]
) -> str:
    """
    Replace all the ? in the query with the actual values from params.

    Args:
        query: The SQL query with placeholders.
        params: A list of parameters to inject into the query.

    Returns:
        str: The query with parameters injected.
    """
    import math

    placeholder_count = query.count("?")
    params_len = len(params) if params else 0

    # Check if the number of placeholders matches the number of parameters
    if placeholder_count != params_len:
        raise ValueError(
            f"The number of placeholders ({placeholder_count}) does not match the number of parameters ({params_len})."
        )

    if not params:
        return query

    # Escape single quotes in parameters to prevent SQL injection
    C1 = "'"
    C2 = "''"
    escaped_params = []
    for param in params:
        if isinstance(param, str):
            new_param = param.replace(C1, C2)
            escaped_params.append(f"'{new_param}'")
        elif isinstance(param, (int, float)):
            # Assert that it's not infinity or NaN
            if not math.isfinite(param):
                raise ValueError(
                    f"Infinity or NaN is not supported as a query parameter. Parameters: {params}"
                )
            v = str(param)
            escaped_params.append(v)
        else:
            raise ValueError(f"Unsupported parameter: {param} type: {type(param)}")

    # Split the query by placeholders and interleave with parameters
    parts = query.split("?")
    query_with_params = "".join(
        part + param for part, param in zip(parts, escaped_params)
    )

    if parts[-1] != "":  # If the last part is not empty, add it too
        query_with_params += parts[-1]

    return query_with_params



import typing
from typing import Optional

if typing.TYPE_CHECKING:
    from sema4ai.actions._action_context import DataContext

    from sema4ai.data._data_server_connection import DataServerConnection
    from sema4ai.data._result_set import ResultSet


class ConnectionNotSetupError(Exception):
    """
    Exception raised when the connection to the data server is not setup.
    """


class _ConnectionHolder:
    _connection: Optional["DataServerConnection"] = None

    @classmethod
    def connection(cls) -> "DataServerConnection":
        if _ConnectionHolder._connection is None:
            raise ConnectionNotSetupError(
                "The connection to the data server is not setup."
            )
        return _ConnectionHolder._connection


class DataSource:
    """
    DataSource refers to a given source configured in the data server.

    i.e.: In the data server, it's possible to configure a postgres connection
    with a name like "my_postgres_datasource". Then, this class can be used
    to refer to that datasource (without having to explicitly add the name of
    the datasource in the written SQL afterwards).
    """

    @property
    def datasource_name(self) -> str:
        """
        The name of the datasource.
        """
        raise NotImplementedError()

    def connection(self) -> "DataServerConnection":
        """
        The connection to the data server.
        """
        return _ConnectionHolder.connection()

    @classmethod
    def setup_connection_from_input_json(cls, value: dict):
        from sema4ai.data._connection_provider import _ConnectionProviderFromDict

        connection_provider = _ConnectionProviderFromDict(value)
        connection = connection_provider.connection()
        _ConnectionHolder._connection = connection

    @classmethod
    def setup_connection_from_data_context(cls, data_context: "DataContext"):
        from sema4ai.data._connection_provider import (
            _ConnectionProviderFromDataContextOrEnvVar,
        )

        connection_provider = _ConnectionProviderFromDataContextOrEnvVar(
            data_context, "data-server"
        )
        connection = connection_provider.connection()
        _ConnectionHolder._connection = connection

    @classmethod
    def setup_connection_from_env_vars(cls):
        from sema4ai.data._connection_provider import (
            _ConnectionProviderFromDataContextOrEnvVar,
        )

        connection_provider = _ConnectionProviderFromDataContextOrEnvVar(None, "")
        connection = connection_provider.connection()
        _ConnectionHolder._connection = connection

    @classmethod
    def model_validate(cls, *, datasource_name: str) -> "DataSource":
        """
        Creates a DataSource given its name.

        Return: A DataSource instance with the given value.

        Note: the model_validate method is used for compatibility with
            the pydantic API.
        """

        return _DataSourceImpl(datasource_name)

    def query(
        self,
        query: str,
        params: Optional[list[str | int | float] | dict[str, str | int | float]] = None,
    ) -> "ResultSet":
        return self.connection().query(query, params)

    def predict(
        self,
        query: str,
        params: Optional[list[str | int | float] | dict[str, str | int | float]] = None,
    ) -> "ResultSet":
        return self.connection().predict(query, params)


class _DataSourceImpl(DataSource):
    """
    Actual implementation of DataSource (not exposed as we can tweak as needed, only the public API should be relied upon).
    """

    def __init__(self, datasource_name: str):
        self._datasource_name = datasource_name

    @property
    def datasource_name(self) -> str:
        return self._datasource_name
